# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2024 Nvidia, Inc. All rights reserved. See COPYING file

#cython: language_level=3

from pyverbs.mr cimport DmaBufMR

cdef class Mlx5DmaBufMR(DmaBufMR):
    pass



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2024 Nvidia, Inc. All rights reserved. See COPYING file

cimport pyverbs.providers.mlx5.libmlx5 as dv
from pyverbs.base import PyverbsRDMAErrno
from pyverbs.mr cimport DmaBufMR
from pyverbs.pd cimport PD


cdef class Mlx5DmaBufMR(DmaBufMR):
    def __init__(self, PD pd not None, offset, length, iova=0, fd=None,
                 access=0, mlx5_access=0):
        """
        Initializes a DmaBufMR (DMA-BUF Memory Region) of the given length
        and access flags using the given PD and DmaBuf objects.
        :param pd: A PD object
        :param offset: Byte offset from the beginning of the dma-buf
        :param length: Length in bytes
        :param iova: The virtual base address of the MR when accessed through a lkey or rkey.
        :param fd: FD representing a dmabuf.
        :param access: Access flags, see ibv_access_flags enum.
        :param mlx5_access: A specific device access flags.
        :return: The newly created DMABUFMR
        """
        self.mr = dv.mlx5dv_reg_dmabuf_mr(pd.pd, offset, length, iova, fd, access, mlx5_access)
        if self.mr == NULL:
            raise PyverbsRDMAErrno(
                f'Failed to register a mlx5 dma-buf MR. length: {length}, access flags: {access} '
                f'mlx5_access: {mlx5_access}')
        self.pd = pd
        self.dmabuf = fd
        self.offset = offset
        pd.add_ref(self)



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2021 Nvidia Inc. All rights reserved. See COPYING file

"""
Test module for mlx5 DevX.
"""

from tests.mlx5_base import Mlx5DevxRcResources, Mlx5DevxTrafficBase
import pyverbs.mem_alloc as mem
from pyverbs.mr import MR
import pyverbs.enums as e
import tests.utils as u


class Mlx5DevxRcOdpRes(Mlx5DevxRcResources):
    @u.requires_odpv2
    def create_mr(self):
        self.with_odp = True
        self.user_addr = mem.mmap(length=self.msg_size,
                                  flags=mem.MAP_ANONYMOUS_ | mem.MAP_PRIVATE_)
        access = e.IBV_ACCESS_LOCAL_WRITE | e.IBV_ACCESS_REMOTE_READ | \
                 e.IBV_ACCESS_ON_DEMAND
        self.mr = MR(self.pd, self.msg_size, access, self.user_addr)


class Mlx5DevxRcTrafficTest(Mlx5DevxTrafficBase):
    """
    Test various functionality of mlx5 DevX objects
    """

    def test_devx_rc_qp_send_imm_traffic(self):
        """
        Creates two DevX RC QPs and modifies them to RTS state.
        Then does SEND_IMM traffic.
        """
        self.create_players(Mlx5DevxRcResources)
        # Send traffic
        self.send_imm_traffic()

    def test_devx_rc_qp_send_imm_doorbell_less_traffic(self):
        """
        Creates two DevX RC QPs with dbr less ext and modifies them to RTS state.
        Then does SEND_IMM traffic.
        """
        from tests.mlx5_prm_structs import SendDbrMode

        self.create_players(Mlx5DevxRcResources, send_dbr_mode=SendDbrMode.NO_DBR_EXT)
        # Send traffic
        self.send_imm_traffic()

    @u.requires_odp('rc', e.IBV_ODP_SUPPORT_SEND | e.IBV_ODP_SUPPORT_RECV)
    def test_devx_rc_qp_odp_traffic(self):
        """
        Creates two DevX RC QPs using ODP enabled MKeys.
        Then does SEND_IMM traffic.
        """
        self.create_players(Mlx5DevxRcOdpRes)
        # Send traffic
        self.send_imm_traffic()



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2024 Nvidia Inc. All rights reserved. See COPYING file

from os import strerror
import unittest
import errno

from pyverbs.providers.mlx5.mlx5dv import Mlx5Context, Mlx5DVContextAttr
from pyverbs.providers.mlx5.mlx5dv_dmabuf import Mlx5DmaBufMR
from pyverbs.pyverbs_error import PyverbsRDMAError
import pyverbs.providers.mlx5.mlx5_enums as dve
from tests.mlx5_base import Mlx5RDMATestCase
from tests.base import RCResources
from pyverbs.qp import QPAttr
import tests.cuda_utils as cu
import pyverbs.enums as e
import tests.utils as u

try:
    from cuda import cuda
    cu.CUDA_FOUND = True
except ImportError:
    cu.CUDA_FOUND = False


GPU_PAGE_SIZE = 1 << 16


def requires_data_direct_support():
    """
    Check if the device support data-direct
    """
    def outer(func):
        def inner(instance):
            with Mlx5Context(Mlx5DVContextAttr(), name=instance.dev_name) as ctx:
                try:
                    ctx.get_data_direct_sysfs_path()
                except PyverbsRDMAError as ex:
                    if ex.error_code == errno.ENODEV:
                        raise unittest.SkipTest('There is no data direct device in the system')
                    raise ex
            return func(instance)
        return inner
    return outer


@cu.set_mem_io_cuda_methods
class Mlx5DmabufCudaRes(RCResources):
    def __init__(self, dev_name, ib_port, gid_index,
                 mr_access=e.IBV_ACCESS_LOCAL_WRITE, mlx5_access=0):
        """
        Initializes data-direct MR and DMA BUF resources on top of a CUDA memory.
        Uses RC QPs for traffic.
        :param dev_name: Device name to be used
        :param ib_port: IB port of the device to use
        :param gid_index: Which GID index to use
        :param mr_access: The MR access
        :param mlx5_access: The data-direct access
        """
        self.mr_access = mr_access
        self.mlx5_access = mlx5_access
        self.cuda_addr = None
        super().__init__(dev_name=dev_name, ib_port=ib_port, gid_index=gid_index)

    def create_mr(self):
        self.cuda_addr = cu.check_cuda_errors(cuda.cuMemAlloc(GPU_PAGE_SIZE))

        attr_flag = 1
        cu.check_cuda_errors(cuda.cuPointerSetAttribute(
            attr_flag,
            cuda.CUpointer_attribute.CU_POINTER_ATTRIBUTE_SYNC_MEMOPS,
            int(self.cuda_addr)))

        cuda_flag = cuda.CUmemRangeHandleType.CU_MEM_RANGE_HANDLE_TYPE_DMA_BUF_FD
        dmabuf_fd = cu.check_cuda_errors(
            cuda.cuMemGetHandleForAddressRange(self.cuda_addr,
                                               GPU_PAGE_SIZE,
                                               cuda_flag,
                                               0))
        try:
            self.mr = Mlx5DmaBufMR(self.pd, offset=0, length=self.msg_size, access=self.mr_access,
                                   fd=dmabuf_fd, mlx5_access=self.mlx5_access)
        except PyverbsRDMAError as ex:
            if ex.error_code == errno.EOPNOTSUPP:
                raise unittest.SkipTest('Registering DV DMABUF MR is not supported')
            raise ex

    def create_qp_attr(self):
        qp_attr = QPAttr(port_num=self.ib_port)
        qp_access = e.IBV_ACCESS_LOCAL_WRITE | e.IBV_ACCESS_REMOTE_WRITE | \
                    e.IBV_ACCESS_REMOTE_READ
        qp_attr.qp_access_flags = qp_access
        return qp_attr


@cu.set_init_cuda_methods
class Mlx5DmabufCudaTest(Mlx5RDMATestCase):
    """
    Test data-direct DV verbs
    """
    @requires_data_direct_support()
    def test_data_direct_sysfs_path_bad_length(self):
        """
        Query data direct sysfs path with buffer of 5 bytes. This is bad flow since 5 bytes aren't
        enough for any sysfs path, so ENOSPC should be raised.
        """
        ctx = Mlx5Context(Mlx5DVContextAttr(), name=self.dev_name)
        try:
            path = ctx.get_data_direct_sysfs_path(5)
        except PyverbsRDMAError as ex:
            self.assertEqual(ex.error_code, errno.ENOSPC,
                             f'Got {strerror(ex.error_code)} but Expected {strerror(errno.ENOSPC)}')
        else:
            raise PyverbsRDMAError('Successfully queried data direct sysfs path with 5 bytes: '
                                   f'{path}')

    def test_dv_dmabuf_mr(self):
        """
        Creates dmabuf MR with DV API. mlx5_access is 0, so the MR is regular dmabuf MR.
        Run RDMA write traffic.
        """
        access = e.IBV_ACCESS_LOCAL_WRITE | e.IBV_ACCESS_REMOTE_WRITE
        self.create_players(Mlx5DmabufCudaRes, mr_access=access)
        u.rdma_traffic(**self.traffic_args, send_op=e.IBV_WR_RDMA_WRITE)

    @requires_data_direct_support()
    def test_dv_dmabuf_mr_data_direct(self):
        """
        Runs RDMA Write traffic over CUDA allocated memory using Data Direct DMA BUF and
        RC QPs.
        """
        mr_access = e.IBV_ACCESS_LOCAL_WRITE | e.IBV_ACCESS_REMOTE_WRITE
        self.create_players(Mlx5DmabufCudaRes, mr_access=mr_access,
                            mlx5_access=dve.MLX5DV_REG_DMABUF_ACCESS_DATA_DIRECT_)
        u.rdma_traffic(**self.traffic_args, send_op=e.IBV_WR_RDMA_WRITE)



# SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
# Copyright (c) 2024 NVIDIA Corporation . All rights reserved. See COPYING file

import unittest
import random
import errno

from pyverbs.providers.mlx5.mlx5dv import Mlx5Context, Mlx5DVContextAttr, \
    Mlx5DVQPInitAttr, Mlx5QP, Mlx5DVCQInitAttr, Mlx5CQ
from pyverbs.pyverbs_error import PyverbsRDMAError, PyverbsUserError, \
    PyverbsError
from pyverbs.cq import CQ, CQEX, PollCqAttr, CqInitAttrEx
from pyverbs.qp import QPInitAttrEx, QPCap, QPAttr
import pyverbs.providers.mlx5.mlx5_enums as dve
from pyverbs.wr import SGE, SendWR, RecvWR
from pyverbs.mr import MR
import pyverbs.enums as e

from tests.base import RCResources, RDMATestCase
from tests.mlx5_base import Mlx5RcResources
import tests.utils as u


def create_ooo_dv_qp(res, max_recv_wr=1000, qp_type=e.IBV_QPT_RC):
    dv_ctx = res.ctx.query_mlx5_device()
    if not dv_ctx.comp_mask & dve.MLX5DV_CONTEXT_MASK_OOO_RECV_WRS:
        raise unittest.SkipTest('DV QP OOO feature is not supported')
    send_ops_flags = e.IBV_QP_EX_WITH_SEND | e.IBV_QP_EX_WITH_SEND_WITH_IMM | \
                     e.IBV_QP_EX_WITH_RDMA_WRITE | e.IBV_QP_EX_WITH_RDMA_READ |\
                     e.IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM
    qp_cap = QPCap(max_recv_wr=max_recv_wr, max_send_wr=max_recv_wr)
    comp_mask = e.IBV_QP_INIT_ATTR_PD | e.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS
    qp_init_attr =  QPInitAttrEx(cap=qp_cap, pd=res.pd, scq=res.cq,
                                 rcq=res.cq, qp_type=qp_type,
                                 send_ops_flags=send_ops_flags,
                                 comp_mask=comp_mask)
    dv_comp_mask = dve.MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS
    attr = Mlx5DVQPInitAttr(comp_mask=dv_comp_mask,
                            create_flags=res.dvqp_create_flags)
    try:
        qp = Mlx5QP(res.ctx, qp_init_attr, attr)
        res.qps.append(qp)
        res.qps_num.append(qp.qp_num)
        res.psns.append(random.getrandbits(24))
    except PyverbsRDMAError as ex:
        raise ex


class Mlx5OOORcRes(Mlx5RcResources):
    def __init__(self, dev_name, ib_port, gid_index, msg_size=1024, dvqp_create_flags=0, **kwargs):
        """
        Initialize mlx5 DV QP resources based on RCResources.
        :param dev_name: Device name to be used
        :param ib_port: IB port of the device to use
        :param gid_index: Which GID index to use
        :param msg_size: The resource msg size
        :param dvqp_create_flags: DV QP create flags
        :param kwargs: General arguments
        """
        self.qp_access_flags = e.IBV_ACCESS_LOCAL_WRITE | e.IBV_ACCESS_REMOTE_WRITE | \
                               e.IBV_ACCESS_REMOTE_READ
        self.dvqp_create_flags = dvqp_create_flags
        super().__init__(dev_name, ib_port, gid_index, msg_size=msg_size, **kwargs)


    def create_qp_attr(self):
        attr = super().create_qp_attr()
        attr.qp_access_flags = self.qp_access_flags
        return attr

    def create_qps(self):
        for _ in range(self.qp_count):
            create_ooo_dv_qp(self)

    def create_mr(self):
        self.mr = MR(self.pd, self.msg_size, self.qp_access_flags)

    def create_cq(self):
        wc_flags = e.IBV_WC_STANDARD_FLAGS
        cia = CqInitAttrEx(cqe=2000, wc_flags=wc_flags)
        dvcq_init_attr = Mlx5DVCQInitAttr()
        dvcq_init_attr.comp_mask |= dve.MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE
        dvcq_init_attr.cqe_size = 64
        dvcq_init_attr.comp_mask |= dve.MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE
        try:
            self.cq = Mlx5CQ(self.ctx, cia, dvcq_init_attr)
        except PyverbsRDMAError as ex:
            if ex.error_code == errno.EOPNOTSUPP:
                raise unittest.SkipTest('Create Mlx5DV CQ is not supported')
            raise ex


class DvOOOQPTest(RDMATestCase):

    def test_ooo_qp_bad_flow(self):
        """
        DDP - OOO Recv WRs bad flow test
        1. Create QP with max recv WRs possible and validate it with querying the QP
        2. Try to create QP with more then max recv wr supported
        3. Try to create QP with unsupported QP type
        """
        self.create_players(Mlx5OOORcRes, dvqp_create_flags=dve.MLX5DV_QP_CREATE_OOO_DP)
        dv_ctx = self.server.ctx.query_mlx5_device()
        max_rc_rwrs = dv_ctx.ooo_recv_wrs_caps['max_rc']
        create_ooo_dv_qp(self.server, max_recv_wr=max_rc_rwrs)
        attr, init_attr = self.server.qps[-1].query(0x1ffffff)
        self.assertEqual(max_rc_rwrs, init_attr.cap.max_recv_wr)
        # Try to create QP with more then max recv wr supported
        with self.assertRaises(PyverbsRDMAError) as ex:
            create_ooo_dv_qp(self.server, max_rc_rwrs + 1)
        self.assertEqual(ex.exception.error_code, errno.EINVAL)
        # Try to create QP with unsupported QP type
        with self.assertRaises(PyverbsRDMAError) as ex:
            create_ooo_dv_qp(self.server, qp_type=e.IBV_QPT_RAW_PACKET)
        self.assertEqual(ex.exception.error_code, errno.EOPNOTSUPP)

    def test_ooo_qp_send_traffic(self):
        """
        DV QP OOO traffic opcode SEND
        """
        self.create_players(Mlx5OOORcRes, dvqp_create_flags=dve.MLX5DV_QP_CREATE_OOO_DP)
        u.traffic_poll_at_once(self, msg_size=int(self.server.msg_size / self.iters),
                               iterations=self.iters)

    def test_ooo_qp_rdma_write_imm_traffic(self):
        """
        DV QP OOO traffic opcode RDMA_WRITE_WITH_IMM
        """
        self.create_players(Mlx5OOORcRes, dvqp_create_flags=dve.MLX5DV_QP_CREATE_OOO_DP)
        u.traffic_poll_at_once(self, msg_size=int(self.server.msg_size / self.iters),
                               iterations=self.iters, opcode=e.IBV_WR_RDMA_WRITE_WITH_IMM)
